"""remove audio_format field.

Revision ID: 880c5a2d80ed
Revises: bf49ac6805f7
Create Date: 2022-11-01 13:41:12.407819

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "880c5a2d80ed"
down_revision = "bf49ac6805f7"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("track", schema=None) as batch_op:
        batch_op.drop_column("audio_format")


def downgrade():
    with op.batch_alter_table("track", schema=None) as batch_op:
        batch_op.add_column(sa.Column("audio_format", sa.VARCHAR(), nullable=True))



"""Core interface for editing an item."""

import datetime
import logging

from sqlalchemy import Date, Integer

from moe.library import LibItem
from moe.library.lib_item import SetType

__all__ = ["EditError", "edit_item"]

log = logging.getLogger("moe.edit")


class EditError(Exception):
    """Error editing an item in the library."""


def edit_item(item: LibItem, field: str, value: str):  # noqa: C901
    """Sets a LibItem's ``field`` to ``value``.

    Args:
        item: Library item to edit.
        field: Item field to edit.
        value: Value to set the item's field to.

    Raises:
        EditError: ``field`` is not a valid attribute or is not editable.
    """
    log.debug(f"Editing item. [{item=!r}, {field=!r}, {value=!r}]")

    if field == "path":
        raise EditError(f"Non-editable field given. [{field=!r}]")

    try:
        attr = getattr(item.__class__, field)
    except AttributeError as a_err:
        if field in item._custom_fields:
            setattr(item, field, value)
            return

        raise EditError(f"Invalid field given. [{field=!r}]") from a_err

    try:
        column_type = attr.property.columns[0].type
    except AttributeError:
        # hybrid_property
        setattr(item, field, value)
        return

    if isinstance(column_type, Integer):
        setattr(item, field, int(value))
    elif isinstance(column_type, SetType):
        setattr(item, field, {value.strip() for value in value.split(";")})
    elif isinstance(column_type, Date):
        try:
            setattr(item, field, datetime.date.fromisoformat(value))
        except ValueError as v_err:
            raise EditError("Date must be in format YYYY-MM-DD") from v_err
    else:
        setattr(item, field, value)

    log.info(f"Item edited. [{item=!r}, {field=!r}, {value=!r}]")



"""json multi value fields.

Revision ID: bf49ac6805f7
Revises: ab11c34c1d0b
Create Date: 2022-10-17 21:09:09.265544

"""
import sqlalchemy as sa
from sqlalchemy.ext.mutable import MutableSet
from sqlalchemy.orm import Session, declarative_base

from alembic import op
from moe.library.lib_item import SetType

# revision identifiers, used by Alembic.
revision = "bf49ac6805f7"
down_revision = "ab11c34c1d0b"
branch_labels = None
depends_on = None

Base = declarative_base()


class _Artist(Base):
    """A track can have multiple artists."""

    __tablename__ = "artist"

    _id = sa.Column(sa.Integer, primary_key=True)
    _track_id = sa.Column(sa.Integer, sa.ForeignKey("track._id"))
    name = sa.Column(sa.String, nullable=False)

    def __init__(self, name: str):
        self.name = name


class _Genre(Base):
    """A track can have multiple genres."""

    __tablename__ = "genre"

    _id = sa.Column(sa.Integer, primary_key=True)
    _track_id = sa.Column(sa.Integer, sa.ForeignKey("track._id"))
    name = sa.Column(sa.String, nullable=False)

    def __init__(self, name: str):
        self.name = name


class _CatalogNums(Base):
    """An album can have multiple catalog numbers."""

    __tablename__ = "catalog_num"

    _id = sa.Column(sa.Integer, primary_key=True)
    _album_id = sa.Column(sa.Integer, sa.ForeignKey("album._id"))
    catalog_num = sa.Column(sa.String, nullable=False)

    def __init__(self, catalog_num: str):
        self.catalog_num = catalog_num


class Track(Base):
    __tablename__ = "track"

    _id = sa.Column(sa.Integer, primary_key=True)
    artists = MutableSet.as_mutable(sa.Column(SetType))
    genres = MutableSet.as_mutable(sa.Column(SetType))


class Album(Base):
    __tablename__ = "album"

    _id = sa.Column(sa.Integer, primary_key=True)
    catalog_nums = MutableSet.as_mutable(sa.Column(SetType))


def upgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("catalog_nums", MutableSet.as_mutable(SetType()), nullable=True)
        )

    with op.batch_alter_table("track", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("artists", MutableSet.as_mutable(SetType()), nullable=True)
        )
        batch_op.add_column(
            sa.Column("genres", MutableSet.as_mutable(SetType()), nullable=True)
        )

    with Session(bind=op.get_bind()) as session:
        tracks = session.query(Track).all()
        for track in tracks:
            genres = session.query(_Genre).filter(_Genre._track_id == track._id).all()
            track.genres = {genre.name for genre in genres}

            artists = (
                session.query(_Artist).filter(_Artist._track_id == track._id).all()
            )
            track.artists = {artist.name for artist in artists}

        albums = session.query(Album).all()
        for album in albums:
            catalog_nums = (
                session.query(_CatalogNums)
                .filter(_CatalogNums._album_id == album._id)
                .all()
            )
            album.catalog_nums = {
                catalog_num.catalog_num for catalog_num in catalog_nums
            }
        session.commit()

    op.drop_table("genre")
    op.drop_table("catalog_num")
    op.drop_table("artist")


def downgrade():
    op.create_table(
        "artist",
        sa.Column("_id", sa.INTEGER(), nullable=False),
        sa.Column("_track_id", sa.INTEGER(), nullable=True),
        sa.Column("name", sa.VARCHAR(), nullable=False),
        sa.ForeignKeyConstraint(
            ["_track_id"],
            ["track._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
    )
    op.create_table(
        "catalog_num",
        sa.Column("_id", sa.INTEGER(), nullable=False),
        sa.Column("_album_id", sa.INTEGER(), nullable=True),
        sa.Column("catalog_num", sa.VARCHAR(), nullable=False),
        sa.ForeignKeyConstraint(
            ["_album_id"],
            ["album._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
    )
    op.create_table(
        "genre",
        sa.Column("_id", sa.INTEGER(), nullable=False),
        sa.Column("_track_id", sa.INTEGER(), nullable=True),
        sa.Column("name", sa.VARCHAR(), nullable=False),
        sa.ForeignKeyConstraint(
            ["_track_id"],
            ["track._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
    )

    with Session(bind=op.get_bind()) as session:
        tracks = session.query(Track).all()
        for track in tracks:
            genres = track.genres
            if genres is not None:
                for genre in genres:
                    db_genre = _Genre(genre)
                    db_genre._track_id = track._id
                    session.add(db_genre)

            artists = track.artists
            if artists is not None:
                for artist in artists:
                    db_artist = _Artist(artist)
                    db_artist._track_id = track._id
                    session.add(db_artist)

        albums = session.query(Album).all()
        for album in albums:
            catalog_nums = album.catalog_nums
            if catalog_nums is not None:
                for catalog_num in catalog_nums:
                    db_catalog_num = _CatalogNums(catalog_num)
                    db_catalog_num._album_id = album._id
                    session.add(db_catalog_num)

        session.commit()

    with op.batch_alter_table("track", schema=None) as batch_op:
        batch_op.drop_column("genres")
        batch_op.drop_column("artists")

    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.drop_column("catalog_nums")



"""new field audio_format.

Revision ID: ab11c34c1d0b
Revises: 0ce5960a081e
Create Date: 2022-10-13 15:55:35.303129

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "ab11c34c1d0b"
down_revision = "0ce5960a081e"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("track", schema=None) as batch_op:
        batch_op.add_column(sa.Column("audio_format", sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table("track", schema=None) as batch_op:
        batch_op.drop_column("audio_format")



"""new field catalog_nums.

Revision ID: 0ce5960a081e
Revises: 32e9fea590b7
Create Date: 2022-10-12 16:13:33.297005

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "0ce5960a081e"
down_revision = "32e9fea590b7"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "catalog_num",
        sa.Column("_id", sa.Integer(), nullable=False),
        sa.Column("_album_id", sa.Integer(), nullable=True),
        sa.Column("catalog_num", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["_album_id"],
            ["album._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
    )


def downgrade():
    op.drop_table("catalog_num")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['moe',
 'moe.library',
 'moe.plugins',
 'moe.plugins.add',
 'moe.plugins.duplicate',
 'moe.plugins.edit',
 'moe.plugins.moe_import',
 'moe.plugins.move',
 'moe.plugins.musicbrainz',
 'moe.plugins.read',
 'moe.plugins.remove',
 'moe.plugins.sync',
 'moe.util',
 'moe.util.cli',
 'moe.util.core']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.15,<2.0.0',
 'Unidecode>=1.2.0,<2.0.0',
 'alembic>=1.4.2,<2.0.0',
 'dynaconf>=3.1.4,<4.0.0',
 'moe_mediafile>=0.10.1,<0.11.0',
 'musicbrainzngs>=0.7.1,<0.8.0',
 'pluggy>=0.13.1,<0.14.0',
 'pyyaml>=5.3.1,<6.0.0',
 'questionary>=1.9.0,<2.0.0',
 'rich>=12.5.1,<13.0.0']

entry_points = \
{'console_scripts': ['moe = moe.cli:main']}

setup_kwargs = {
    'name': 'moe',
    'version': '1.5.1',
    'description': 'The ultimate tool for managing your music library.',
    'long_description': '###############\nWelcome to Moe!\n###############\nMoe is our resident Music-Organizer-Extraordinaire who\'s sole purpose is to give you full control over your music library by streamlining the process between downloading/ripping music to your filesystem and listening to it with your favorite music player.\n\nIn short, Moe maintains a database of your music library that can be updated with various metadata sources, queried, edited, etc. through either an API or command-line interface. All of these features, and more, are made available by a highly extensible plugin ecosystem.\n\nUsage\n=====\nMoe comes with a command-line interface which is how most users will take advantage of the library management features. Below is a screenshot of Moe importing an album from the filesystem and adding it to the underlying database all while fixing tags, relocating the files, and anything else you can imagine.\n\n.. image:: _static/import_example.png\n\nAlternatively, because all the core functionality is available via an API, the underlying music management system can be incorporated into any existing program or other user interface.\n\nFinally, although a lot of Moe\'s functionality exists around the idea of operating on a library database of your music, the database is entirely optional. In this case, Moe provides a convenient set of tools and functionality for handling music in a general sense. For example, below is a standalone python script that takes an album directory and Musicbrainz release ID from the command-line, and then updates the underlying files\' tags with any changes from Musicbrainz.\n\n.. code:: python\n\n    #!/usr/bin/env python3\n\n    import argparse\n    import pathlib\n\n    import moe.plugins.musicbrainz as moe_mb\n    from moe.config import Config, ConfigValidationError\n    from moe.library import Album\n    from moe.plugins.write import write_tags\n\n\n    def main():\n        try:\n            Config(init_db=False)\n        except ConfigValidationError as err:\n            raise SystemExit(1) from err\n\n        parser = argparse.ArgumentParser(\n            description="Update an album with musicbrainz tags."\n        )\n        parser.add_argument("path", help="dir of the album to update")\n        parser.add_argument("mb_id", help="musicbrainz id of the album to fetch")\n        args = parser.parse_args()\n\n        album = Album.from_dir(pathlib.Path(args.path))\n\n        album.merge(moe_mb.get_album_by_id(args.mb_id), overwrite=True)\n\n        for track in album.tracks:\n            write_tags(track)\n\n\n    if __name__ == "__main__":\n        main()\n\n.. note::\n\n   Notice the use of ``init_db=False`` when initializing the configuration to tell Moe you don\'t want to use or create a database file.\n\n\nThis is just a small taste of what Moe is capable of and how it can make your life easier when dealing with music in Python. Stop re-inventing the wheel; use Moe.\n\nIf you want to learn more, check out the `Getting Started <https://mrmoe.readthedocs.io/en/latest/getting_started.html>`_ docs.\n',
    'author': 'Jacob Pavlock',
    'author_email': 'jtpavlock@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/MoeMusic/Moe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)
